import caterpie
import abc
from abc import ABC
import pandas as pd

class Table(ABC):

    def __init__(self, source, types_dict, name = '', **kwargs):

        self.load(source, **kwargs)
        self.types_dict = types_dict
        self.table_name = name

    def preprocess(self): pass # Can perform operations before writing

    def postprocess(self): pass # Can perform cleanup operations after writing

    @abc.abstractmethod
    def load(self, source): pass

    def unload(self):
        """ Removes df to clean memory. """
        self.message = None
        self.types_dict = None

    def infer_types(self): pass # Possible for future

class CSV(Table):

    def load(self, message, index_col = None):
        """ Converts csv into dataframe. """
        if type(message) is pd.DataFrame:
            self.message = message # NOTE: Assume argument is df for now.
        elif type(message) is str:
            self.message = pd.read_csv(message, index_col=index_col)
        else:
            raise TypeError("df must be a pandas dataframe or path to csv file on disk.")

class SourceTable(Table):

    def load(self, source):
        self.message = source.read()

class Writer:

    def __init__(self, csvs, conn, backend='postgresql'):

        self.conn = conn
        self.csvs = csvs
        self.backend = set_backend(backend)

    def write_tables(self): # TODO: Should be idempotent and add columns as needed
        """ Loop through csvs and call COPY FROM command to import them. """
        for csv in self.csvs:
            csv.preprocess()
            if not self.backend.table_exists(csv.table_name, self.conn):
                self.backend.create_table(csv.table_name, csv.types_dict, self.conn)
            if not self.table_up_to_date(csv): # TODO: Should check if data is already present
                self.update_table(csv)
            csv.postprocess()

    def update_table(self, csv):

        # Add columns if needed
        alterations = {key: value for key, value in csv.types_dict.items()
                        if not self.backend.column_in_table(csv.table_name, key, self.conn)}
        if alterations:
            print("Adding columns {0} to table {1}".format(alterations, csv.table_name))
            self.backend.add_columns(csv.table_name, alterations, self.conn)

        # Add rows
        self.backend.update_table(csv.table_name, csv.message, self.conn, columns = list(csv.types_dict.keys()))

    def table_up_to_date(self, csv):
        """ Determined if csv needs to be written/updated or if it is already in database. """
        return False

def set_backend(backend):

    if backend == 'postgresql':
        try:
            from caterpie.postgresql import postgresql_utils as pg_backend
        except ImportError as e:
            raise ImportError("In order to use the postgresql backend, you must have psycopg2 installed. \
                You can install using 'pip3 install psycopg2' or 'pip3 install psycopg2-binary' \
                Additional error messages: {error}".format(error=e))
        return pg_backend
    if backend == 'sqlite':
        raise NotImplementedError
        from caterpie.sqlite import sqlite_utils as sqlite_backend
        return sqlite_backend
    else:
        raise ValueError("Backend {backend} does not exist.".format(backend=backend))



import psycopg2 as pg
from caterpie import read_environment

def get_connection(
    host=None,
    dbname=None,
    username=None,
    password=None,
    port=None,
):
    host = host or read_environment('psql_host')
    dbname = dbname or read_environment('dbname')
    username = username or read_environment('username')
    password = password or read_environment('password')
    port = port or read_environment('port')

    conn_string = "user=%s password=%s dbname=%s host=%s port = %s" % (username, password, dbname, host, port)
    return pg.connect(conn_string)


import psycopg2  as pg
from caterpie import postgresql
from caterpie.postgresql import postgresql_utils as pu
import pandas as pd
import random
import names

def make_dummy_table(conn, table_name = 'test'):

    curr = conn.cursor()
    curr.execute("CREATE TABLE {0} ( id INT, name VARCHAR(20), gender CHAR(20), url VARCHAR(80), weight FLOAT );".format(table_name))
    conn.commit()

def make_dummy_data(n):

    dummy = {
            'id': [random.randint(0,10*n) for _ in range(n)],
            'name': [names.get_first_name() for _ in range(n)],
            'gender': [random.choice(['m','f','o']) for _ in range(n)],
            'url': [pu.random_string(5) for _ in range(n)],
            'weight': [random.random()*10*n for _ in range(n)]
            }
    return pd.DataFrame(dummy)

def count_none(iterable):

    counts = [item is None for item in iterable]
    return sum(counts)

def test_table_exists():

    conn = postgresql.get_connection()
    table_name = 'existence'
    pu.drop_table(table_name, conn)
    make_dummy_table(conn, table_name)
    assert pu.table_exists(table_name, conn)
    assert not pu.table_exists('lol', conn)
    pu.drop_table(table_name, conn)

def test_column_in_table():

    conn = postgresql.get_connection()
    table_name = 'columnz'
    pu.drop_table(table_name, conn)
    make_dummy_table(conn, table_name)
    assert pu.column_in_table(table_name, 'gender', conn)
    assert not pu.column_in_table(table_name, 'doggy', conn)
    pu.drop_table(table_name, conn)

def test_get_columns():

    conn = postgresql.get_connection()
    table_name = 'get_column'
    pu.drop_table(table_name, conn)
    make_dummy_table(conn, table_name)
    columns = pu.get_columns(table_name, conn)
    expected = ['id', 'name', 'gender', 'url', 'weight']
    for c in expected:
        assert c in columns
    pu.drop_table(table_name, conn)

def test_create_and_delete():

    conn = postgresql.get_connection()
    types_dict = {
        "id": "INT",
        "name": "VARCHAR(40)"
    }
    table = 'test'
    pu.drop_table(table, conn)
    pu.create_table(table, types_dict, conn)
    assert pu.table_exists(table, conn)
    pu.drop_table(table, conn)
    assert not pu.table_exists(table, conn)

def test_random_string():

    n = 12
    rando = pu.random_string(n)
    assert type(rando) is str
    assert len(rando) == n

def test_add_columns():

    conn = postgresql.get_connection()
    table_name = 'add_column'
    pu.drop_table(table_name, conn)
    make_dummy_table(conn,table_name)
    n = 20
    pu.update_table(table_name, make_dummy_data(n), conn)
    pu.add_columns(table_name, {'donkey': 'INT', 'chocobo': 'VARCHAR(9)'}, conn)
    pu.update_table(table_name, make_dummy_data(n), conn)
    curr = conn.cursor()
    curr.execute("SELECT * FROM {0};".format(table_name))
    results = curr.fetchall()
    assert len(results) == 2*n
    for result in results: #
        assert count_none(result) == 2
    pu.drop_table(table_name, conn)

def test_update_table():

    conn = postgresql.get_connection()
    table_name = 'update'
    pu.drop_table(table_name, conn)
    make_dummy_table(conn, table_name)
    n = 20
    df = make_dummy_data(n)
    pu.update_table(table_name, df, conn)
    pu.update_table(table_name, df, conn)
    pu.update_table(table_name, df, conn, columns = ['name', 'gender'])
    pu.update_table(table_name, df, conn, columns = ['name'])
    curr = conn.cursor()
    curr.execute("SELECT * FROM {0};".format(table_name))
    results = curr.fetchall()
    assert len(results) == 4*n
    for result in results[0:2*n]:
        assert count_none(result) == 0
    for result in results[2*n:3*n]:
        assert count_none(result) == 3
    for result in results[3*n:]:
        assert count_none(result) == 4
    pu.drop_table(table_name, conn)



import psycopg2  as pg
import pandas as pd
import os
import random
import string
from shutil import copyfile
import re

def table_exists(table, conn):
    """ Returns true if requested table exists in database. """
    curr = conn.cursor()
    curr.execute(
        "SELECT EXISTS ("
            "SELECT 1 "
            "FROM pg_tables "
            "WHERE schemaname = 'public' "
            "AND tablename = '{0}' "
            ");".format(table))
    return curr.fetchall()[0][0]

def column_in_table(table, column, conn):
    """ Returns true if requested column is present in requested table. """
    curr = conn.cursor()
    curr.execute(
        "SELECT EXISTS ("
            "SELECT attname "
            "FROM pg_attribute "
            "WHERE attrelid = (SELECT OID FROM pg_class WHERE relname = '{0}') "
            "AND attname = '{1}'"
            ");".format(table, column.lower())
    )
    return curr.fetchall()[0][0]

def get_columns(table, conn):
    """ Returns columns of requested table. """
    curr = conn.cursor()
    curr.execute(
        "SELECT attname "
        "FROM pg_attribute "
        "WHERE attrelid = (SELECT OID FROM pg_class WHERE relname = '{0}') "
        ";".format(table)
    )
    columns = curr.fetchall()
    return [c[0] for c in columns] # Formatting

def create_table(table, types_dict, conn):
    """ Creates table with paramaters specified by the types_dict. """

    print("Creating table {0}".format(table))
    curr = conn.cursor()
    code = "CREATE TABLE {0} (".format(table)
    columns = ["{0} {1}".format(column, types_dict[column]) for column in types_dict]
    code += ", ".join(columns)
    code += ");"
    curr.execute(code)
    conn.commit()

def drop_table(table, conn):
    """ Deletes table if it exists in connected db. """
    print("Dropping table {0}".format(table))
    curr = conn.cursor()
    curr.execute("DROP TABLE IF EXISTS {0}".format(table))
    conn.commit()

def add_columns(table, update_dict, conn):

    print("Adding columns {0} to table {1}".format(str(update_dict), table))
    curr = conn.cursor()
    code = "ALTER TABLE {0} ".format(table)
    updates = ["ADD COLUMN {0} {1}".format(column, column_type) for column, column_type in update_dict.items()]
    code += ", ".join(updates) + ';'
    curr.execute(code)
    conn.commit()

def update_table(table, update_file, conn, columns = None):

    temp = '/tmp/' + random_string() + '.csv' # TODO: Make sure files are unique
    try:
        os.remove(temp)
    except OSError:
        pass
    if type(update_file) is str:
        df = pd.read_csv(update_file, index_col=0)
    elif type(update_file) is pd.DataFrame:
        df = update_file
    else:
        raise TypeError("Update_file must be a path to a csv or a dataframe.")
    if columns is None: # Use all columns
        columns = df.columns

    df[columns].to_csv(temp, index=False, header=True)
    temp_file = open(temp)
    _update_table(table, temp_file, columns, conn) # Save df to temp file and upload
    temp_file.close()
    os.remove(temp)

def _update_table(table, path, columns, conn):

    curr = conn.cursor()
    curr.copy_expert(
        sql = "COPY {0} ({1}) FROM stdin WITH CSV HEADER DELIMITER as ',';".format(table, ', '.join(columns)),
        file = path
        )
    conn.commit()

def random_string(n = 10):
    """ Creates a random string of length n from lower case, upper case, and digits. """
    characters = string.ascii_letters + string.digits
    return ''.join(random.choice(characters) for _ in range(n))

def clean(string):
    """ Replaces illegal characters in a string for postgres. """
    return re.sub('[-()]','_', string)



import psycopg2
from caterpie import postgresql
from caterpie import CSV, Writer
import caterpie.postgresql.postgresql_utils as pu
import pandas as pd


def make_test():

    df = pd.read_csv('test/phenoData')
    df = df.rename(columns = {'Unnamed: 0': 'sampleID'})

    types_dict = {
                'sampleID': 'VARCHAR(32)',
                'subjectID': 'VARCHAR(32)',
                'body_site': 'VARCHAR(30)',
                'antibiotics_current_use': 'BOOLEAN',
                'study_condition': 'VARCHAR(20)',
                'disease': 'VARCHAR(20)',
                'age': 'INT',
                'infant_age': 'INT',
                'age_category': 'VARCHAR(20)',
                'gender': 'VARCHAR(20)',
                'country': 'CHAR(3)',
                'non_westernized': 'BOOLEAN',
                'DNA_extraction_kit': 'VARCHAR(30)',
                'number_reads': 'INT',
                'number_bases': 'INT',
                'minimum_read_length': 'INT',
                'median_read_length': 'INT',
                'pregnant': 'BOOLEAN',
                'lactating': 'BOOLEAN',
                'NCBI_accession': 'CHAR(10)'
        }

    return df, types_dict

def make_other_test():

    df = pd.read_csv('test/trapData')

    types_dict = {
        'Date': 'DATE',
        'Open': 'float',
        'High': 'float',
        'Low': 'float',
        'Close': 'float'
    }
    return df, types_dict


def test_csv():

    df, types_dict = make_test()
    csv = CSV(df, types_dict, name='test')
    assert csv.table_name == 'test'
    for key in types_dict:
        assert csv.types_dict[key] == types_dict[key]
    assert csv.message.equals(df)

def test_update_table():

    conn = postgresql.get_connection()
    curr = conn.cursor()
    table_name = 'update_test'
    pu.drop_table(table_name, conn)
    df, types_dict = make_test()
    csv = CSV(df, types_dict, name=table_name)
    writer = Writer([csv], conn)
    pu.create_table(table_name, types_dict, conn)
    writer.update_table(csv)
    curr.execute("SELECT * FROM {0};".format(table_name))
    results = curr.fetchall()
    assert len(results) == 8
    for result in results:
        assert len(result) == 20


def test_write_tables():

    conn = postgresql.get_connection()
    curr = conn.cursor()
    table_name = 'write_test'
    pu.drop_table(table_name, conn)
    df, types_dict = make_test()
    csv = CSV(df, types_dict, name=table_name)
    df, types_dict = make_other_test()
    nvda = CSV(df, types_dict, name=table_name)
    writer = Writer([csv, nvda], conn)
    writer.write_tables()
    curr.execute("SELECT * FROM {0};".format(table_name))
    results = curr.fetchall()
    for result in results:
        assert len(result) == 25



import configparser
import os
from os import environ as env
from caterpie.table import CSV, Writer
from dotenv import load_dotenv

load_dotenv()

def read_environment(key):
    try:
        return env[key]
    except KeyError as e:
        raise KeyError("Could not infer a value for {0}. \
            You must either set an environment variable called {0} or \
            specify the value of {0} as a keyword argument.".format(key)
        )


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['caterpie', 'caterpie.postgresql']

package_data = \
{'': ['*'], 'caterpie': ['sqlite/*']}

install_requires = \
['configparser>=4.0.2,<5.0.0',
 'pandas>=0.25.3,<0.26.0',
 'python-dotenv>=0.10.3,<0.11.0']

setup_kwargs = {
    'name': 'caterpie',
    'version': '0.1.0',
    'description': 'Lightweight ORM for uploading CSV files to a relational database.',
    'long_description': None,
    'author': 'smk508',
    'author_email': 'skhan8@mail.einstein.yu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
